#include "940.h"

extern int dzzStartVorbisDecode(unsigned char *pData, int nDataLen,
		unsigned char *pResultData, unsigned char *pMallocData);
extern int dzzDecodeVorbisBlock();
extern void dzzEndVorbisDecode();
extern void dzzSetVorbisOutLocation(unsigned char *pResultData);
extern void dzzmemcpy(void *pDest, void *pSrc, int n);
extern char *dzzstrcpy(char *to, const char *from);
extern int dzzstrlen(const char *s);

// variables shared with the 920

SharedVariables *g_pShared;

// variables for the 940 alone

// the music loop output stuff has a 100,000 byte buffer
// located at g_psMusicLoopOutPointer.  This is a circular buffer
// that has both a nNextToReadFrom and a nNextToWriteTo pointer.
#define MUSICOUTBUFSIZE 100000
short *g_psMusicLoopOutPointer;
short *g_psMusicLoopTempBuf;
int g_nNextMusicLoopOutToReadFrom;
int g_nNextMusicLoopOutToWriteTo;
int g_bMusicReady;

int g_bDecodingStarted;

unsigned char *g_pMallocMemory;

static void FillAudioBuffer(short *pS)
{
	int n;

	for(n = 0; n < 4096; n++)
	{
		if(g_pShared->g_bMusicOn && g_bMusicReady && 
				(g_nNextMusicLoopOutToReadFrom != g_nNextMusicLoopOutToWriteTo))
		{
			*pS = g_psMusicLoopOutPointer[g_nNextMusicLoopOutToReadFrom];
			g_nNextMusicLoopOutToReadFrom++;
			if(g_nNextMusicLoopOutToReadFrom == (MUSICOUTBUFSIZE>>1))
				g_nNextMusicLoopOutToReadFrom = 0;
		}
		pS++;
	}
}

void CheckFillBuffers()
{
	if(g_pShared->g_bMusicOn && g_bMusicReady)
	{
		if(!g_pShared->g_bSoundBuffer0Full)
		{
			FillAudioBuffer(g_pShared->g_sSoundBuffer0);
			g_pShared->g_bSoundBuffer0Full = DZZTRUE;
		}
		if(!g_pShared->g_bSoundBuffer1Full)
		{
			FillAudioBuffer(g_pShared->g_sSoundBuffer1);
			g_pShared->g_bSoundBuffer1Full = DZZTRUE;
		}
	}
}

void DoSoundTick()
{
	if(g_pShared->g_bResetMusic)
	{
		g_bMusicReady = 0;
		g_pShared->g_bResetMusic = DZZFALSE;
		g_bDecodingStarted = DZZFALSE;
		g_nNextMusicLoopOutToReadFrom = 0;
		g_nNextMusicLoopOutToWriteTo = 0;
	}

	CheckFillBuffers();

	if(g_pShared->g_bMusicLoaded)
	{
		// cases for various program states:

		// 1. Music is being loaded off of disk so we cannot decode anything.  If
		// that's true, the pShared->g_bMusicLoaded will be false.
		// 2. Music has not yet started.  We need to start the vorbis decode.  There are
		// two cases where that can happen -- either the current decoding is finished
		// or we have loaded a new song.  If we have loaded a new song, the reset which
		// was processed above takes care of setting it, otherwise it gets set at the end
		// of the song
		// 3. We're going along fine and the buffer is less than half full.  In that case,
		// decode a block
		// 4. We're going along fine but the buffer is more than half full.  In that case,
		// set g_bMusicReady but don't do anything else

		int nLen;

		if(!g_bDecodingStarted)
		{
			g_bDecodingStarted = DZZTRUE;
			dzzStartVorbisDecode(g_pShared->g_pMusicLoopInPointer,
				g_pShared->g_nMusicLoopInLength, (unsigned char *)
				g_psMusicLoopTempBuf, g_pMallocMemory);
		}
		else
		{
			int bLessThanHalfFull = 0;
			// check to make sure that the buffer is less than half full
			if(g_nNextMusicLoopOutToReadFrom < g_nNextMusicLoopOutToWriteTo)
			{
				if(g_nNextMusicLoopOutToWriteTo - g_nNextMusicLoopOutToReadFrom < MUSICOUTBUFSIZE/4)
					bLessThanHalfFull = 1;
			}
			else if(g_nNextMusicLoopOutToReadFrom == g_nNextMusicLoopOutToWriteTo)
				bLessThanHalfFull = 1;
			else
			{
				if(g_nNextMusicLoopOutToReadFrom - g_nNextMusicLoopOutToWriteTo > MUSICOUTBUFSIZE/4)
					bLessThanHalfFull = 1;
			}
			if(bLessThanHalfFull)
			{
				nLen = dzzDecodeVorbisBlock();
				if(nLen == 0)
				{
					g_bDecodingStarted = DZZFALSE;
					dzzEndVorbisDecode();
				}
				else
				{
					int n;
					// copy the data from g_psMusicLoopTempBuf to
					// the output loop, then call a function to reset
					// the decode output pointer back to the temp buf
					for(n = 0; n < nLen/2; n++)
					{
						g_psMusicLoopOutPointer[g_nNextMusicLoopOutToWriteTo] = g_psMusicLoopTempBuf[n];
						g_nNextMusicLoopOutToWriteTo++;
						if(g_nNextMusicLoopOutToWriteTo == (MUSICOUTBUFSIZE>>1))
							g_nNextMusicLoopOutToWriteTo = 0;
					}
					dzzSetVorbisOutLocation((unsigned char *) g_psMusicLoopTempBuf);
				}
			}
			else
			{
				g_bMusicReady = 1;
			}
		}
	}
}

void CheckForImportantStuff(char *m)
{
	CheckFillBuffers();
}

void Main940()
{
	SharedVariables *pVars = (SharedVariables *) 0x200000;
	g_bDecodingStarted = DZZFALSE;

	g_pShared = pVars;

	g_pMallocMemory = (unsigned char*) 0x14C000;

	// set up the music stuff
	g_psMusicLoopOutPointer = (short *) (0xE00000);
	g_psMusicLoopTempBuf = (short *) (0x3CC000);
	g_nNextMusicLoopOutToReadFrom = 0;
	g_nNextMusicLoopOutToWriteTo = 0;
	g_bMusicReady = 0;

	while(!g_pShared->g_bQuit)
	{
		DoSoundTick();
	}
}

